use crate::store::object::{Obj, ObjType};
use crate::utils::path::normalize_path;
use std::cmp::Ordering;
use std::fs::File;
use std::io::{self, Write};
use std::io::{BufRead, BufReader};
use std::path::PathBuf;

// Custom sorting function to handle paths hierarchically
fn sort_paths_hierarchically(paths: &mut Vec<PathBuf>) {
    paths.sort_by(|a, b| {
        // Split paths into components for comparison
        let a_components: Vec<_> = a.components().collect();
        let b_components: Vec<_> = b.components().collect();

        // Compare components one by one
        for (a_component, b_component) in a_components.iter().zip(b_components.iter()) {
            match a_component.cmp(b_component) {
                Ordering::Equal => continue,
                ordering => return ordering,
            }
        }

        // If one path is a prefix of the other, the shorter path comes first
        a_components.len().cmp(&b_components.len())
    });
}

#[derive(Debug)]
struct IndexedObj {
    obj_type: ObjType,
    path: PathBuf,
}

pub struct Indexer {
    repo_root: PathBuf,
    index_file: PathBuf,
    indexed_objs: Vec<IndexedObj>,
}

impl Indexer {
    pub fn new(repo_root: &PathBuf) -> Self {
        let mut index_file = repo_root.clone();
        index_file.push(".nextsync");
        index_file.push("index");

        Indexer {
            repo_root: repo_root.clone(),
            index_file,
            indexed_objs: Vec::new(),
        }
    }

    pub fn load(&mut self) -> io::Result<()> {
        let file = File::open(&self.index_file)?;
        let reader = BufReader::new(file);

        for line in reader.lines() {
            let line = line?;
            let line = line.as_bytes();

            let path_str = String::from_utf8(line[1..].to_vec()).unwrap();
            let path = PathBuf::from(path_str);

            self.indexed_objs.push(IndexedObj {
                obj_type: ObjType::try_from(line[0]).unwrap(),
                path,
            });
        }

        Ok(())
    }

    pub fn load_unsafe(&mut self) {
        let _ = self.load();
        dbg!(&self.indexed_objs);
    }

    fn is_indexed(&self, obj: &IndexedObj) -> bool {
        self.indexed_objs
            .iter()
            .position(|o| o.obj_type == obj.obj_type && o.path == obj.path)
            .is_some()
    }

    pub fn is_staged(&self, obj: &Obj) -> bool {
        dbg!(obj);
        // self.indexed_objs.iter().position(|o| &o.obj_type == obj.get_obj_type() && &o.path == obj.get_obj_path()).is_some()
        self.indexed_objs
            .iter()
            .position(|o| &o.path == obj.get_obj_path())
            .is_some()
    }

    pub fn is_staged_parent(&self, obj: &Obj) -> bool {
        self.indexed_objs
            .iter()
            .position(|o| o.path.starts_with(obj.get_obj_path()))
            .is_some()
    }

    fn index_obj(&mut self, path: PathBuf, obj_type: ObjType) {
        let mut path = path;
        path = normalize_path(path); // normalize path (/foo/./bar => /foo/bar)
        // change path to be relative to repo's root
        path = path.strip_prefix(&self.repo_root).unwrap().to_path_buf();

        let indexed_obj = IndexedObj {
            obj_type,
            path: path.clone(),
        };

        if self.is_indexed(&indexed_obj) {
            return;
        }

        self.indexed_objs.push(indexed_obj);
    }

    pub fn index_file(&mut self, path: PathBuf) {
        self.index_obj(path, ObjType::Blob);
    }

    pub fn index_dir(&mut self, path: PathBuf) {
        self.index_obj(path, ObjType::Tree);
    }

    pub fn save(&self) -> io::Result<()> {
        let mut file = File::create(&self.index_file)?;

        for obj in self.indexed_objs.iter() {
            // write obj_type
            file.write_all(&[obj.obj_type.clone().into()])?;

            // write path
            file.write_all(obj.path.to_str().unwrap().as_bytes())?;

            file.write_all(b"\n")?;
        }

        Ok(())
    }
}
