use crate::store::object::{Obj, ObjMetadata, ObjType};
use crypto::digest::Digest;
use crypto::sha1::Sha1;
use std::path::PathBuf;
use std::sync::OnceLock;

pub static REPO_ROOT: OnceLock<PathBuf> = OnceLock::new();

pub fn init(repo_root: &PathBuf) {
    REPO_ROOT.set(repo_root.clone());
}

type NsObjectChilds = Vec<Box<NsObject>>;

struct NsObjectPath {
    path: PathBuf,
}

impl From<&str> for NsObjectPath {
    fn from(hash: &str) -> Self {
        let (dir, res) = hash.split_at(2);

        let mut ns_obj_path = match REPO_ROOT.get() {
            Some(path) => path.clone(),
            None => {
                panic!("fatal: 'REPO_ROOT' not set, you must initialize nsobject before using it!")
            }
        };
        ns_obj_path.push(dir);
        ns_obj_path.push(res);

        NsObjectPath { path: ns_obj_path }
    }
}

impl From<&PathBuf> for NsObjectPath {
    fn from(obj_path: &PathBuf) -> Self {
        let mut hasher = Sha1::new();
        hasher.input_str(
            obj_path
                .to_str()
                .expect("Cannot contains non UTF-8 char in path"),
        );
        NsObjectPath::from(hasher.result_str().as_str())
    }
}

pub struct NsObject {
    pub obj_type: ObjType,
    obj_path: OnceLock<PathBuf>,
    nsobj_path: OnceLock<PathBuf>,
    childs: OnceLock<NsObjectChilds>,
    index: usize,
}

impl NsObject {
    pub fn from_local_path(path: &PathBuf) -> Self {
        NsObject {
            obj_type: ObjType::Obj,
            obj_path: OnceLock::from(path.to_path_buf()),
            nsobj_path: OnceLock::new(),
            childs: OnceLock::new(),
            index: 0,
        }
    }

    pub fn from_hash(hash: &str) -> Self {
        NsObject {
            obj_type: ObjType::Obj,
            obj_path: OnceLock::new(),
            nsobj_path: OnceLock::from(NsObjectPath::from(hash).path),
            childs: OnceLock::new(),
            index: 0,
        }
    }

    pub fn get_obj_path(&self) -> &PathBuf {
        self.obj_path.get_or_init(|| todo!())
    }

    fn get_nsobj_path(&self) -> &PathBuf {
        self.nsobj_path
            .get_or_init(|| NsObjectPath::from(self.get_obj_path()).path)
    }

    /// Return the existence of the nsobj in the store
    pub fn exists(&self) -> bool {
        println!("{}", self.get_nsobj_path().display());
        self.get_nsobj_path().exists()
    }

    pub fn get_childs(&self) -> &NsObjectChilds {
        self.childs.get_or_init(|| {
            if self.obj_type != ObjType::Tree {
                Vec::new()
            } else {
                todo!()
            }
        })
    }

    pub fn is_file(&self) -> bool {
        self.obj_type == ObjType::Blob
    }

    pub fn is_dir(&self) -> bool {
        self.obj_type == ObjType::Tree
    }

    pub fn get_local_obj(&self) -> Obj {
        let mut obj = Obj::from_local_path(self.get_obj_path());
        obj.set_type(self.obj_type.clone());
        obj
    }

    /// Saving the nsobj in the store
    ///
    /// first line consists of all size fixed datas (obj_type, size, modified)
    /// next lines are the size variable datas
    ///
    /// * if it is a Tree obj after an empty line there will be the definition
    /// of its subobjs (one line by subobj) *
    /// obj_type + hash
    pub fn save(&self) -> Result<(), ()> {
        if !self.get_obj_path().exists() {
            // delete current obj
            // delete reference on parent
        } else {
        }
        Ok(())
    }

    pub fn get_metadata(&self) -> Option<ObjMetadata> {
        todo!()
    }

    pub fn iter(&self) -> NsObjectIter<'_> {
        NsObjectIter {
            nsobject: self,
            index: 0,
        }
    }
}

pub struct NsObjectIter<'a> {
    nsobject: &'a NsObject,
    index: usize,
}

impl<'a> Iterator for NsObjectIter<'a> {
    type Item = &'a NsObject;

    fn next(&mut self) -> Option<Self::Item> {
        self.index += 1;
        match self.nsobject.get_childs().get(self.index - 1) {
            None => None,
            Some(obj) => Some(&**obj),
        }
    }
}
