use crate::store::nsobject::NsObject;
use crate::utils::path;
use std::fs;
use std::path::PathBuf;
use std::sync::OnceLock;
use std::time::SystemTime;

const MAX_SIZE_TO_USE_HASH: u64 = 12 * 1024 * 1024;

pub struct ObjMetadata {
    size: u64,
    modified: Option<SystemTime>,
}

#[derive(PartialEq, Clone, Debug)]
#[repr(u8)]
pub enum ObjType {
    Obj,
    Blob,
    Tree,
}

impl From<ObjType> for u8 {
    fn from(variant: ObjType) -> Self {
        variant as u8
    }
}


impl TryFrom<u8> for ObjType {
    type Error = String;

    fn try_from(byte: u8) -> Result<Self, Self::Error> {
        match byte {
            0 => Ok(ObjType::Obj),
            1 => Ok(ObjType::Blob),
            2 => Ok(ObjType::Tree),
            _ => Err(format!("Invalid byte value: {}", byte)),
        }
    }
}

#[derive(PartialEq, Clone, Debug)]
pub enum ObjStatus {
    Undefined,
    Created,
    Modified,
    Moved,
    Copied,
    Deleted,
}

#[derive(Clone, Debug)]
pub struct Obj {
    obj_type: ObjType,
    status: OnceLock<ObjStatus>,
    /// path of the object from root
    obj_path: PathBuf,
}

impl Obj {
    pub fn from_local_path(path: &PathBuf) -> Self {
        // todo set state
        Obj {
            obj_type: ObjType::Obj,
            status: OnceLock::new(),
            obj_path: path.to_path_buf(),
        }
    }

    pub fn get_status(&self) -> &ObjStatus {
        self.status.get_or_init(|| {
            // todo!();
            // TODO read path
            ObjStatus::Created
        })
    }

    pub fn set_status(&mut self, status: ObjStatus) {
        self.status = OnceLock::from(status);
    }

    pub fn set_type(&mut self, obj_type: ObjType) {
        self.obj_type = obj_type;
    }

    pub fn is_new(&self) -> bool {
        self.get_status() == &ObjStatus::Created
    }

    pub fn get_obj_type(&self) -> &ObjType {
        &self.obj_type
    }

    pub fn get_obj_path(&self) -> &PathBuf {
        &self.obj_path
    }

    pub fn cpy_path(&self) -> String {
        path::to_string(&self.obj_path)
    }

    pub fn get_metadata(&self) -> Option<ObjMetadata> {
        let metadata = match fs::metadata(&self.obj_path) {
            Ok(m) => m,
            Err(err) => {
                eprintln!(
                    "Failed to read metadata of {} ({})",
                    self.obj_path.display(),
                    err
                );
                return None;
            }
        };

        Some(ObjMetadata {
            size: metadata.len(),
            modified: metadata.modified().ok(),
        })
    }
}

impl PartialEq<NsObject> for Obj {
    fn eq(&self, other: &NsObject) -> bool {
        if self.obj_type != other.obj_type {
            eprintln!("Trying to compare different obj type");
            return false;
        }

        if self.obj_type == ObjType::Tree {
            panic!("fatal: cannot comparing folders");
        }

        let obj_metadata = self.get_metadata();
        let nsobj_metadata = other.get_metadata();

        // If not metadata exist considere it is not the same
        if obj_metadata.is_none() || nsobj_metadata.is_none() {
            return false;
        }

        let obj_metadata = obj_metadata.unwrap();
        let nsobj_metadata = nsobj_metadata.unwrap();

        // if we have modified and it is the same it must be the same file
        if obj_metadata.modified.is_some() && nsobj_metadata.modified.is_some() {
            if obj_metadata.modified.unwrap() == nsobj_metadata.modified.unwrap() {
                return true;
            }
        }

        // if size is same considere comparing hash if file is small
        if obj_metadata.size == obj_metadata.size {
            if obj_metadata.size < MAX_SIZE_TO_USE_HASH
                && nsobj_metadata.size < MAX_SIZE_TO_USE_HASH
            {
                todo!()
                // return self.get_file_hash() == other.get_file_hash();
            } else {
                return true;
            }
        } else {
            return false;
        }
    }
}
